@@ -6,13 +6,8 @@
 #include <sstream>
 #include <iomanip>
 #include <string>
-
-#ifndef HGWRR_RCPP
+#include <utility>
 #include <armadillo>
-#else
-#include <RcppArmadillo.h>
-#endif
-
 #include <gsl/gsl_multimin.h>
 #include <gsl/gsl_errno.h>
 
@@ -33,8 +28,6 @@ struct ML_Params
     uword q;
 };
 
-typedef vec (*GWRKernelFunctionSquared)(vec, double);
-
 inline vec gwr_kernel_gaussian2(vec dist2, double bw2)
 {
     return exp(- dist2 / (2.0 * bw2));
@@ -50,6 +43,91 @@ const GWRKernelFunctionSquared gwr_kernel_functions[] = {
     gwr_kernel_bisquare2
 };
 
+double criterion_bw(const double bw, const HLMGWRBWArgs* args, const bool verbose, const PrintFunction pcout)
+{
+    const mat G = args->G;
+    const mat Vig = args->Vig;
+    const vec Viy = args->Viy;
+    const mat u = args->u;
+    const size_t ngroup = Viy.n_rows;
+    /// Calibrate for each gorup.
+    GWRKernelFunctionSquared gwr_kernel = args->gwr_kernel;
+    double cv = 0.0;
+    for (size_t i = 0; i < ngroup; i++)
+    {
+        mat d_u = u.each_row() - u.row(i);
+        vec d2 = sum(d_u % d_u, 1);
+        double b2 = vec(sort(d2))[(int)bw - 1];
+        vec wW = (*gwr_kernel)(d2, b2);
+        wW(i) = 0;
+        mat GtWVG = (G.each_col() % wW).t() * Vig;
+        mat GtWVy = (G.each_col() % wW).t() * Viy;
+        try
+        {
+            vec bi = solve(GtWVG, GtWVy);
+            double yhi = as_scalar(Vig.row(i) * bi);
+            double residual = Viy(i) - yhi;
+            cv += residual * residual;
+        }
+        catch(const std::exception& e)
+        {
+            return DBL_MAX;
+        }
+    }
+    if (verbose)
+    {
+        ostringstream sout;
+        sout << "bw: " << bw << "; " << "cv: " << cv << "\r";
+        pcout(sout.str());
+    }
+    return cv;
+}
+
+double golden_selection(const double lower, const double upper, const bool adaptive, const HLMGWRBWArgs* args, const bool verbose, const PrintFunction pcout)
+{
+    double xU = upper, xL = lower;
+    bool adaptBw = adaptive;
+    const double eps = 1e-4;
+    const double R = (sqrt(5)-1)/2;
+    int iter = 0;
+    double d = R * (xU - xL);
+    double x1 = adaptBw ? floor(xL + d) : (xL + d);
+    double x2 = adaptBw ? round(xU - d) : (xU - d);
+    double f1 = criterion_bw(x1, args, verbose, pcout);
+    double f2 = criterion_bw(x2, args, verbose, pcout);
+    double d1 = f2 - f1;
+    double xopt = f1 < f2 ? x1 : x2;
+    double ea = 100;
+    while ((fabs(d) > eps) && (fabs(d1) > eps) && iter < ea)
+    {
+        d = R * d;
+        if (f1 < f2)
+        {
+            xL = x2;
+            x2 = x1;
+            x1 = adaptBw ? round(xL + d) : (xL + d);
+            f2 = f1;
+            f1 = criterion_bw(x1, args, verbose, pcout);
+        }
+        else
+        {
+            xU = x1;
+            x1 = x2;
+            x2 = adaptBw ? floor(xU - d) : (xU - d);
+            f1 = f2;
+            f2 = criterion_bw(x2, args, verbose, pcout);
+        }
+        iter = iter + 1;
+        xopt = (f1 < f2) ? x1 : x2;
+        d1 = f2 - f1;
+    }
+    if (verbose)
+    {
+        pcout("\n");
+    }
+    return xopt;
+}
+
 /**
  * @brief Estimate $\gamma$.
  * 
@@ -62,7 +140,11 @@ const GWRKernelFunctionSquared gwr_kernel_functions[] = {
  * @param wD Equals to $D$
  * @return mat 
  */
-mat fit_gwr(const mat& G, const vec* Yf, const mat* Zf, const size_t ngroup, const mat& D, const mat& u, const double bw, const GWRKernelType kernel)
+tuple<mat, double> fit_gwr(
+    const mat& G, const vec* Yf, const mat* Zf, const size_t ngroup,
+    const mat& D, const mat& u, const double bw, const GWRKernelType kernel,
+    const size_t verbose, const PrintFunction pcout
+)
 {
     uword k = G.n_cols;//, q = Zf[0].n_cols;
     mat beta(ngroup, k, arma::fill::zeros), D_inv = D.i();
@@ -78,19 +160,27 @@ mat fit_gwr(const mat& G, const vec* Yf, const mat* Zf, const size_t ngroup, con
         Vig.row(i) = Visigma * ones(ndata, 1) * G.row(i);
         Viy(i) = as_scalar(Visigma * Yi);
     }
-    /// Calibrate for each gorup.
     GWRKernelFunctionSquared gwr_kernel = gwr_kernel_functions[kernel];
+    /// Check whether need to optimize bw
+    double b = bw;
+    if (bw == 0.0)
+    {
+        HLMGWRBWArgs bw_args { G, Vig, Viy, u, gwr_kernel };
+        double upper = ngroup, lower = k + 1;
+        b = golden_selection(lower, upper, true, &bw_args, verbose > 1, pcout);
+    }
+    /// Calibrate for each gorup.
     for (size_t i = 0; i < ngroup; i++)
     {
         mat d_u = u.each_row() - u.row(i);
         vec d2 = sum(d_u % d_u, 1);
-        double b2 = vec(sort(d2))[(int)bw];
+        double b2 = vec(sort(d2))[(int)b - 1];
         vec wW = (*gwr_kernel)(d2, b2);
-        mat GtWVG = G.t() * (Vig.each_col() % wW);
-        mat GtWVy = G.t() * (Viy % wW);
+        mat GtWVG = (G.each_col() % wW).t() * Vig;
+        mat GtWVy = (G.each_col() % wW).t() * Viy;
         beta.row(i) = solve(GtWVG, GtWVy).t();
     }
-    return beta;
+    return make_tuple(beta, b);
 }
 
 vec fit_gls(const mat* Xf, const vec* Yf, const mat* Zf, const size_t ngroup, const mat& D)
@@ -203,7 +293,8 @@ double ml_gsl_f_D(const gsl_vector* v, void* p)
     }
     mat D(q, q, arma::fill::zeros);
     D(trimatl_ind(size(D))) = D_tri;
-    D(trimatu_ind(size(D))) = D_tri;
+    D = D.t();
+    D(trimatl_ind(size(D))) = D_tri;
     double logL = loglikelihood(Xf, Yf, Zf, ngroup, D, *beta, n);
     return -logL / double(n);
 }
@@ -230,7 +321,8 @@ double ml_gsl_f_D_beta(const gsl_vector* v, void* pparams)
     }
     mat D(q, q, arma::fill::zeros);
     D(trimatl_ind(size(D))) = D_tri;
-    D(trimatu_ind(size(D))) = D_tri;
+    D = D.t();
+    D(trimatl_ind(size(D))) = D_tri;
     double logL = loglikelihood(Xf, Yf, Zf, ngroup, D, beta, n);
     return -logL / double(n);
 }
@@ -253,7 +345,8 @@ void ml_gsl_df_D(const gsl_vector* v, void* p, gsl_vector *df)
     }
     mat D(q, q, arma::fill::zeros);
     D(trimatl_ind(size(D))) = D_tri;
-    D(trimatu_ind(size(D))) = D_tri;
+    D = D.t();
+    D(trimatl_ind(size(D))) = D_tri;
     mat dL_D;
     loglikelihood_d(Xf, Yf, Zf, ngroup, D, *beta, n, dL_D);
     dL_D = -dL_D / double(n);
@@ -286,7 +379,8 @@ void ml_gsl_df_D_beta(const gsl_vector* v, void* pparams, gsl_vector *df)
     }
     mat D(q, q, arma::fill::zeros);
     D(trimatl_ind(size(D))) = D_tri;
-    D(trimatu_ind(size(D))) = D_tri;
+    D = D.t();
+    D(trimatl_ind(size(D))) = D_tri;
     mat dL_D;
     vec dL_beta;
     loglikelihood_d(Xf, Yf, Zf, ngroup, D, beta, n, dL_D, dL_beta);
@@ -315,7 +409,7 @@ void ml_gsl_fdf_D_beta(const gsl_vector* v, void* p, double *f, gsl_vector *df)
     ml_gsl_df_D_beta(v, p, df);
 }
 
-void fit_D(mat& D, const ML_Params* params, const double alpha, const double eps, const size_t max_iters, const bool verbose, const PrintFunction pcout)
+double fit_D(mat& D, const ML_Params* params, const double alpha, const double eps, const size_t max_iters, const bool verbose, const PrintFunction pcout)
 {
     int precision = int(log10(1.0 / eps));
     uword q = D.n_cols, ntarget = q * (q + 1) / 2;
@@ -340,8 +434,18 @@ void fit_D(mat& D, const ML_Params* params, const double alpha, const double eps
     if (verbose)
     {
         ostringstream sout;
-        sout << setprecision(precision) << fixed << minimizer->x->data[0] << "," << minimizer->x->data[1] << "," << minimizer->x->data[2] << ";";
-        sout << setprecision(precision) << fixed << minimizer->gradient->data[0] << "," << minimizer->gradient->data[1] << "," << minimizer->gradient->data[2] << ";";
+        sout << setprecision(precision) << fixed << minimizer->x->data[0];
+        for (size_t i = 1; i < D_tril_vec.n_elem; i++)
+        {
+            sout << "," << minimizer->x->data[i];
+        }
+        sout << ";";
+        sout << setprecision(precision) << fixed << minimizer->gradient->data[0] << ",";
+        for (size_t i = 1; i < D_tril_vec.n_elem; i++)
+        {
+            sout << "," << minimizer->gradient->data[i];
+        }
+        sout << ";";
         sout << minimizer->f << '\r';
         pcout(sout.str());
     }
@@ -354,28 +458,42 @@ void fit_D(mat& D, const ML_Params* params, const double alpha, const double eps
         if (verbose)
         {
             ostringstream sout;
-            sout << setprecision(precision) << fixed << minimizer->x->data[0] << "," << minimizer->x->data[1] << "," << minimizer->x->data[2] << ";";
-            sout << setprecision(precision) << fixed << minimizer->gradient->data[0] << "," << minimizer->gradient->data[1] << "," << minimizer->gradient->data[2] << ";";
+            sout << setprecision(precision) << fixed << minimizer->x->data[0];
+            for (size_t i = 1; i < D_tril_vec.n_elem; i++)
+            {
+                sout << "," << minimizer->x->data[i];
+            }
+            sout << ";";
+            sout << setprecision(precision) << fixed << minimizer->gradient->data[0];
+            for (size_t i = 1; i < D_tril_vec.n_elem; i++)
+            {
+                sout << "," << minimizer->gradient->data[i];
+            }
+            sout << ";";
             sout << minimizer->f << '\r';
             pcout(sout.str());
         }
-        if (status) break;
-        if (minimizer->f < 0 || gsl_isnan(minimizer->f)) break;
+        if (status || gsl_isnan(minimizer->f)) break;
         status = gsl_multimin_test_gradient(minimizer->gradient, eps);
     } while (status == GSL_CONTINUE && (++iter) < max_iters);
     if (verbose) 
     {
         pcout("\n");
     }
-    vec D_tri(arma::size(D_tril_idx));
-    for (uword i = 0; i < ntarget; i++)
+    if (!gsl_isnan(minimizer->f))
     {
-        D_tri(i) = gsl_vector_get(x0, i);
+        mat D1 = mat(arma::size(D), arma::fill::eye);
+        vec D_tri(arma::size(D_tril_idx));
+        for (uword i = 0; i < ntarget; i++)
+        {
+            D_tri(i) = gsl_vector_get(minimizer->x, i);
+        }
+        D1(D_tril_idx) = D_tri;
+        D1 = D1.t();
+        D1(D_tril_idx) = D_tri;
+        D = D1;
     }
-    mat D1 = mat(arma::size(D), arma::fill::zeros);
-    D1(D_tril_idx) = D_tri;
-    D1(D_triu_idx) = D_tri;
-    D = D1;
+    return minimizer->f;
 }
 
 void fit_D_beta(mat& D, vec& beta, const ML_Params* params, const double alpha, const double eps, const size_t max_iters, const bool verbose, const PrintFunction pcout)
@@ -407,12 +525,28 @@ void fit_D_beta(mat& D, vec& beta, const ML_Params* params, const double alpha,
     if (verbose)
     {
         ostringstream sout;
-        sout << setprecision(precision) << fixed << 
-            minimizer->x->data[0] << "," << minimizer->x->data[1] << "," << 
-            minimizer->x->data[2] << "," << minimizer->x->data[3] << "," << minimizer->x->data[4] << ";";
-        sout << setprecision(precision) << fixed << 
-            minimizer->gradient->data[0] << "," << minimizer->gradient->data[1] << "," << 
-            minimizer->gradient->data[2] << "," << minimizer->gradient->data[3] << "," << minimizer->gradient->data[4] << ";";
+        sout << setprecision(precision) << fixed;
+        for (size_t i = 0; i < p; i++)
+        {
+            sout << minimizer->x->data[i] << ",";
+        }
+        sout << minimizer->x->data[p];
+        for (size_t i = 1; i < D_tril_vec.n_elem; i++)
+        {
+            sout << "," << minimizer->x->data[p + i];
+        }
+        sout << ";";
+        sout << setprecision(precision) << fixed;
+        for (size_t i = 0; i < p; i++)
+        {
+            sout << minimizer->gradient->data[i] << ",";
+        }
+        sout << minimizer->x->data[p];
+        for (size_t i = 1; i < D_tril_vec.n_elem; i++)
+        {
+            sout << "," << minimizer->gradient->data[p + i];
+        }
+        sout << ";"; 
         sout << minimizer->f << '\r';
         pcout(sout.str());
     }
@@ -425,35 +559,54 @@ void fit_D_beta(mat& D, vec& beta, const ML_Params* params, const double alpha,
         if (verbose)
         {
             ostringstream sout;
-            sout << setprecision(precision) << fixed << 
-                minimizer->x->data[0] << "," << minimizer->x->data[1] << "," << 
-                minimizer->x->data[2] << "," << minimizer->x->data[3] << "," << minimizer->x->data[4] << ";";
-            sout << setprecision(precision) << fixed << 
-                minimizer->gradient->data[0] << "," << minimizer->gradient->data[1] << "," << 
-                minimizer->gradient->data[2] << "," << minimizer->gradient->data[3] << "," << minimizer->gradient->data[4] << ";";
+            sout << setprecision(precision) << fixed;
+            for (size_t i = 0; i < p; i++)
+            {
+                sout << minimizer->x->data[i] << ",";
+            }
+            sout << minimizer->x->data[p];
+            for (size_t i = 1; i < D_tril_vec.n_elem; i++)
+            {
+                sout << "," << minimizer->x->data[p + i];
+            }
+            sout << ";";
+            sout << setprecision(precision) << fixed;
+            for (size_t i = 0; i < p; i++)
+            {
+                sout << minimizer->gradient->data[i] << ",";
+            }
+            sout << minimizer->x->data[p];
+            for (size_t i = 1; i < D_tril_vec.n_elem; i++)
+            {
+                sout << "," << minimizer->gradient->data[p + i];
+            }
+            sout << ";"; 
             sout << minimizer->f << '\r';
             pcout(sout.str());
         }
-        if (status) break;
-        if (minimizer->f < 0 || gsl_isnan(minimizer->f)) break;
+        if (status || gsl_isnan(minimizer->f)) break;
         status = gsl_multimin_test_gradient(minimizer->gradient, eps);
     } while (status == GSL_CONTINUE && (++iter) < max_iters);
     if (verbose) 
     {
         pcout("\n");
     }
-    vec D_tri(arma::size(D_tril_idx));
-    for (uword i = p; i < ntarget; i++)
+    mat D1(arma::size(D), arma::fill::eye);
+    vec beta1(arma::size(beta), arma::fill::ones);
+    if (!gsl_isnan(minimizer->f))
     {
-        D_tri(i - p) = gsl_vector_get(minimizer->x, i);
-    }
-    mat D1(arma::size(D), arma::fill::zeros);
-    D1(D_tril_idx) = D_tri;
-    D1(D_triu_idx) = D_tri;
-    vec beta1(arma::size(beta), arma::fill::zeros);
-    for (uword i = 0; i < p; i++)
-    {
-        beta1(i) = gsl_vector_get(minimizer->x, i);
+        vec D_tri(arma::size(D_tril_idx));
+        for (uword i = p; i < ntarget; i++)
+        {
+            D_tri(i - p) = gsl_vector_get(minimizer->x, i);
+        }
+        D1(D_tril_idx) = D_tri;
+        D1 = D1.t();
+        D1(D_tril_idx) = D_tri;
+        for (uword i = 0; i < p; i++)
+        {
+            beta1(i) = gsl_vector_get(minimizer->x, i);
+        }
     }
     D = D1;
     beta = beta1;
@@ -517,7 +670,7 @@ HLMGWRParams backfitting_maximum_likelihood(const HLMGWRArgs& args, const HLMGWR
     const vec& y = args.y;
     const mat& u = args.u;
     const uvec& group = args.group;
-    double bw = args.bw;
+    double bw = args.bw, bw_optim = bw;
     GWRKernelType kernel = args.kernel;
     uword ngroup = G.n_rows, ndata = X.n_rows;
     uword nvg = G.n_cols, nvx = X.n_cols, nvz = Z.n_cols;
@@ -551,8 +704,8 @@ HLMGWRParams backfitting_maximum_likelihood(const HLMGWRArgs& args, const HLMGWR
     // Backfitting
     //============
     size_t retry = 0;
-    double rss = 0.0, rss0 = 0.0, diff = DBL_MAX;
-    for (size_t iter = 0; iter < max_iters && diff > eps_iter && retry <= max_retries; iter++)
+    double rss = DBL_MAX, rss0 = DBL_MAX, diff = DBL_MAX, mlf = 0.0;
+    for (size_t iter = 0; (abs(diff) > eps_iter) && iter < max_iters && retry < max_retries; iter++)
     {
         rss0 = rss;
         //--------------------
@@ -562,7 +715,9 @@ HLMGWRParams backfitting_maximum_likelihood(const HLMGWRArgs& args, const HLMGWR
         {
             Ygf[i] = Yf[i] - Xf[i] * beta;
         }
-        gamma = fit_gwr(G, Ygf, Zf, ngroup, D, u, bw, kernel);
+        auto gwr_result = fit_gwr(G, Ygf, Zf, ngroup, D, u, bw, kernel, verbose, pcout);
+        gamma = get<0>(gwr_result);
+        bw_optim = get<1>(gwr_result);
         vec hatMg = sum(G % gamma, 1);
         vec hatM = hatMg.rows(group);
         vec yh = y - hatM;
@@ -577,13 +732,11 @@ HLMGWRParams backfitting_maximum_likelihood(const HLMGWRArgs& args, const HLMGWR
         switch (ml_type)
         {
         case 0:
-            D = eye(size(D));
-            fit_D(D, &ml_params, alpha, eps_gradient, max_iters, verbose > 1, pcout);
+            mlf = fit_D(D, &ml_params, alpha, eps_gradient, max_iters, verbose > 1, pcout);
             beta = fit_gls(Xf, Yhf, Zf, ngroup, D);
             break;
         case 1:
             ml_params.beta = nullptr;
-            D = eye(size(D));
             beta = fit_gls(Xf, Yhf, Zf, ngroup, D);
             fit_D_beta(D, beta, &ml_params, alpha, eps_gradient, max_iters, verbose > 1, pcout);
             break;
@@ -599,18 +752,23 @@ HLMGWRParams backfitting_maximum_likelihood(const HLMGWRArgs& args, const HLMGWR
         vec yhat = yh - (X * beta) - sum(Z % (mu.rows(group)), 1);
         vec residual = yhat % yhat;
         rss = sum(residual);
-        diff = abs(rss - rss0);
-        if (rss > rss0 && iter > 0) retry++;
-        else if (retry > 0) retry = 0;
+        diff = rss - rss0;
+        if (rss < rss0) 
+        {
+            if (retry > 0) retry = 0;
+        }
+        else if (iter > 0) retry++;
         if (verbose > 0)
         {
             ostringstream sout;
-            sout << fixed << setprecision(prescition) <<
-                "RSS: " << rss << ", " <<
-                "diff: " << diff << ", " <<
-                "R2: " << (1 - rss / tss) << ", " <<
-                "Retry: " << retry <<
-                endl;
+            sout << fixed << setprecision(prescition) << "Iter: " << iter;
+            if (bw == 0.0) sout << ", " << "Bw: " << bw_optim;
+            sout << ", " << "RSS: " << rss;
+            if (abs(diff) < DBL_MAX) sout << ", " << "dRSS: " << diff;
+            sout << ", " << "R2: " << (1 - rss / tss);
+            sout << ", " << "-loglik/n: " << mlf;
+            if (retry > 0) sout << ", " << "Retry: " << retry;
+            sout << endl;
             pcout(sout.str());
         }
     }
@@ -620,5 +778,5 @@ HLMGWRParams backfitting_maximum_likelihood(const HLMGWRArgs& args, const HLMGWR
     delete[] Yf;
     delete[] Ygf;
     delete[] Yhf ;
-    return { gamma, beta, mu, D, sigma };
+    return { gamma, beta, mu, D, sigma, bw_optim };
 }
